@@ -11,14 +11,15 @@ There are four main types of documentation objects:
 - `Function`
 - `Variable`
 
-All documentation types types make heavy use of `@functools.cached_property` decorators.
+All documentation types make heavy use of `@functools.cached_property` decorators.
 This means they have a large set of attributes that are lazily computed on first access.
-By convention, all attributes are read-only, although this this not enforced at runtime.
+By convention, all attributes are read-only, although this is not enforced at runtime.
 """
 from __future__ import annotations
 
 import enum
 import inspect
+import os
 import pkgutil
 import re
 import sys
@@ -27,19 +28,21 @@ import traceback
 import types
 import warnings
 from abc import ABCMeta, abstractmethod
+from collections.abc import Callable
 from functools import wraps
 from pathlib import Path
-from typing import Any, ClassVar, Generic, Optional, TypeVar, Union
+from typing import Any, ClassVar, Generic, TypeVar, Union
 
-from pdoc import doc_ast, extract
+from pdoc import doc_ast, doc_pyi, extract
 from pdoc.doc_types import (
+    GenericAlias,
     NonUserDefinedCallables,
     empty,
-    formatannotation,
     resolve_annotations,
     safe_eval_type,
 )
-from ._compat import _tuplegetter, cache, cached_property, get_origin
+
+from ._compat import cache, cached_property, formatannotation, get_origin
 
 
 def _include_fullname_in_traceback(f):
@@ -96,7 +99,7 @@ class Doc(Generic[T]):
     """
     `(modulename, qualname)` of this doc object's original location.
     In the context of a module, this points to the location it was imported from,
-    in the context classes, this points to the class an attribute is inherited from.
+    in the context of classes, this points to the class an attribute is inherited from.
     """
 
     def __init__(
@@ -144,7 +147,7 @@ class Doc(Generic[T]):
         return doc_ast.get_source(self.obj)
 
     @cached_property
-    def source_file(self) -> Optional[Path]:
+    def source_file(self) -> Path | None:
         """The name of the Python source file in which this object was defined. `None` for built-in objects."""
         try:
             return Path(inspect.getsourcefile(self.obj) or inspect.getfile(self.obj))  # type: ignore
@@ -152,7 +155,7 @@ class Doc(Generic[T]):
             return None
 
     @cached_property
-    def source_lines(self) -> Optional[tuple[int, int]]:
+    def source_lines(self) -> tuple[int, int] | None:
         """
         Return a `(start, end)` line number tuple for this object.
 
@@ -212,15 +215,15 @@ class Namespace(Doc[T], metaclass=ABCMeta):
     def _var_docstrings(self) -> dict[str, str]:
         """A mapping from some member variable names to their docstrings."""
 
-    @abstractmethod
-    def _taken_from(self, member_name: str, obj: Any) -> tuple[str, str]:
-        """The location this member was taken from. If unknown, (modulename, qualname) is returned."""
-
     @cached_property
     @abstractmethod
     def _var_annotations(self) -> dict[str, Any]:
         """A mapping from some member variable names to their type annotations."""
 
+    @abstractmethod
+    def _taken_from(self, member_name: str, obj: Any) -> tuple[str, str]:
+        """The location this member was taken from. If unknown, (modulename, qualname) is returned."""
+
     @cached_property
     @abstractmethod
     def own_members(self) -> list[Doc]:
@@ -228,7 +231,11 @@ class Namespace(Doc[T], metaclass=ABCMeta):
 
     @cached_property
     def members(self) -> dict[str, Doc]:
-        """A mapping from all members to their documentation objects."""
+        """A mapping from all members to their documentation objects.
+
+        This mapping includes private members; they are only filtered out as part of the template logic.
+
+        """
         members: dict[str, Doc] = {}
         for name, obj in self._member_objects.items():
             qualname = f"{self.qualname}.{name}".lstrip(".")
@@ -263,26 +270,46 @@ class Namespace(Doc[T], metaclass=ABCMeta):
                     taken_from=taken_from,
                 )
             elif inspect.isroutine(obj):
-                doc = Function(self.modulename, qualname, obj, taken_from)
-            elif inspect.isclass(obj) and obj is not empty:
+                doc = Function(self.modulename, qualname, obj, taken_from)  # type: ignore
+            elif (
+                inspect.isclass(obj)
+                and obj is not empty
+                and not isinstance(obj, GenericAlias)
+            ):
+                # `dict[str,str]` is a GenericAlias instance. We want to render type aliases as variables though.
                 doc = Class(self.modulename, qualname, obj, taken_from)
             elif inspect.ismodule(obj):
-                doc = Module(obj)
-                doc.modulename = self.modulename
-                doc.qualname = qualname
-            else:
-                docstring = self._var_docstrings.get(name, "")
-                if not docstring and inspect.isdatadescriptor(obj):
-                    docstring = getattr(obj, "__doc__", None) or ""
+                if os.environ.get("PDOC_SUBMODULES"):  # pragma: no cover
+                    doc = Module.from_name(obj.__name__)
+                else:
+                    continue
+            elif inspect.isdatadescriptor(obj):
                 doc = Variable(
                     self.modulename,
                     qualname,
-                    docstring=docstring,
+                    docstring=getattr(obj, "__doc__", None) or "",
+                    annotation=self._var_annotations.get(name, empty),
+                    default_value=empty,
+                    taken_from=taken_from,
+                )
+            else:
+                doc = Variable(
+                    self.modulename,
+                    qualname,
+                    docstring="",
                     annotation=self._var_annotations.get(name, empty),
                     default_value=obj,
                     taken_from=taken_from,
                 )
+            if self._var_docstrings.get(name):
+                doc.docstring = self._var_docstrings[name]
             members[doc.name] = doc
+
+        if isinstance(self, Module):
+            # quirk: doc_pyi expects .members to be set already
+            self.members = members  # type: ignore
+            doc_pyi.include_typeinfo_from_stub_files(self)
+
         return members
 
     @cached_property
@@ -320,7 +347,7 @@ class Namespace(Doc[T], metaclass=ABCMeta):
         return flattened
 
     @cache
-    def get(self, identifier: str) -> Optional[Doc]:
+    def get(self, identifier: str) -> Doc | None:
         """Returns the documentation object for a particular identifier, or `None` if the identifier cannot be found."""
         head, _, tail = identifier.partition(".")
         if tail:
@@ -347,6 +374,12 @@ class Module(Namespace[types.ModuleType]):
         """
         super().__init__(module.__name__, "", module, (module.__name__, ""))
 
+    @classmethod
+    @cache
+    def from_name(cls, name: str) -> Module:
+        """Create a `Module` object by supplying the module's (full) name."""
+        return cls(extract.load_module(name))
+
     @cache
     @_include_fullname_in_traceback
     def __repr__(self):
@@ -357,7 +390,7 @@ class Module(Namespace[types.ModuleType]):
         """
         `True` if the module is a package, `False` otherwise.
 
-        Packages are special kind of module that may have submodules.
+        Packages are a special kind of module that may have submodules.
         Typically, this means that this file is in a directory named like the
         module with the name `__init__.py`.
         """
@@ -367,6 +400,14 @@ class Module(Namespace[types.ModuleType]):
     def _var_docstrings(self) -> dict[str, str]:
         return doc_ast.walk_tree(self.obj).docstrings
 
+    @cached_property
+    def _var_annotations(self) -> dict[str, Any]:
+        annotations = doc_ast.walk_tree(self.obj).annotations.copy()
+        for k, v in _safe_getattr(self.obj, "__annotations__", {}).items():
+            annotations[k] = v
+
+        return resolve_annotations(annotations, self.obj, self.fullname)
+
     def _taken_from(self, member_name: str, obj: Any) -> tuple[str, str]:
         if obj is empty:
             return self.modulename, f"{self.qualname}.{member_name}".lstrip(".")
@@ -383,48 +424,46 @@ class Module(Namespace[types.ModuleType]):
                 "."
             )
 
-    @cached_property
-    def _var_annotations(self) -> dict[str, Any]:
-        annotations = doc_ast.walk_tree(self.obj).annotations.copy()
-        for k, v in _safe_getattr(self.obj, "__annotations__", {}).items():
-            annotations[k] = v
-
-        return resolve_annotations(annotations, self.obj, self.fullname)
-
     @cached_property
     def own_members(self) -> list[Doc]:
         return list(self.members.values())
 
     @cached_property
-    def submodules(self) -> list["Module"]:
+    def submodules(self) -> list[Module]:
         """A list of all (direct) submodules."""
         if not self.is_package:
             return []
 
-        if _safe_getattr(self.obj, "__all__", False):
-            # If __all__ is set, only show submodules specified there.
-            return [
-                mod
-                for mod in self.members.values()
-                if isinstance(mod, Module)
-                and mod.modulename.startswith(self.modulename)
-            ]
-
+        include: Callable[[str], bool]
+        mod_all = _safe_getattr(self.obj, "__all__", False)
+        if mod_all is not False:
+            mod_all_pos = {name: i for i, name in enumerate(mod_all)}
+            include = mod_all_pos.__contains__
         else:
-            submodules = []
-            for mod in pkgutil.iter_modules(self.obj.__path__, f"{self.fullname}."):  # type: ignore
-                if mod.name.split(".")[-1].startswith("_"):
-                    continue
-                try:
-                    module = extract.load_module(mod.name)
-                except RuntimeError:
-                    warnings.warn(
-                        f"Couldn't import {mod.name}:\n{traceback.format_exc()}",
-                        RuntimeWarning,
-                    )
-                    continue
-                submodules.append(Module(module))
-            return submodules
+
+            def include(name: str) -> bool:
+                # optimization: we don't even try to load modules starting with an underscore as they would not be
+                # visible by default. The downside of this is that someone who overrides `is_public` will miss those
+                # entries, the upsides are 1) better performance and 2) less warnings because of import failures
+                # (think of OS-specific modules, e.g. _linux.py failing to import on Windows).
+                return not name.startswith("_")
+
+        submodules = []
+        for mod in pkgutil.iter_modules(self.obj.__path__, f"{self.fullname}."):
+            _, _, mod_name = mod.name.rpartition(".")
+            if not include(mod_name):
+                continue
+            try:
+                module = Module.from_name(mod.name)
+            except RuntimeError:
+                warnings.warn(f"Couldn't import {mod.name}:\n{traceback.format_exc()}")
+                continue
+            submodules.append(module)
+
+        if mod_all:
+            submodules = sorted(submodules, key=lambda m: mod_all_pos[m.name])
+
+        return submodules
 
     @cached_property
     def _documented_members(self) -> set[str]:
@@ -434,9 +473,12 @@ class Module(Namespace[types.ModuleType]):
     def _member_objects(self) -> dict[str, Any]:
         members = {}
 
-        all = _safe_getattr(self.obj, "__all__", False)
-        if all:
-            for name in all:
+        all_: list[str] | None = _safe_getattr(self.obj, "__all__", None)
+        if all_ is not None:
+            for name in all_:
+                if not isinstance(name, str):
+                    # Gracefully handle the case where objects are directly specified in __all__.
+                    name = _safe_getattr(name, "__name__", str(name))
                 if name in self.obj.__dict__:
                     val = self.obj.__dict__[name]
                 elif name in self._var_annotations:
@@ -448,27 +490,28 @@ class Module(Namespace[types.ModuleType]):
                         val = extract.load_module(f"{self.modulename}.{name}")
                     except RuntimeError as e:
                         warnings.warn(
-                            f"Found {name!r} in {self.modulename}.__all__, but it does not resolve: {e}",
-                            RuntimeWarning,
+                            f"Found {name!r} in {self.modulename}.__all__, but it does not resolve: {e}"
                         )
                         val = empty
                 members[name] = val
 
         else:
             # Starting with Python 3.10, __annotations__ is created on demand,
-            # we access it here first so that we don't change obj.__dict__ while iterating over it.
-            _safe_getattr(self.obj, "__annotations__", {})
-            for name, obj in self.obj.__dict__.items():
+            # so we make a copy here as obj.__dict__ is changed while we iterate over it.
+            # Additionally, accessing self._documented_members may lead to the execution of TYPE_CHECKING blocks,
+            # which may also modify obj.__dict__. (https://github.com/mitmproxy/pdoc/issues/351)
+            for name, obj in list(self.obj.__dict__.items()):
                 # We already exclude everything here that is imported, only a TypeVar,
                 # or a variable without annotation and docstring.
                 # If one needs to document one of these things, __all__ is the correct way.
-                if isinstance(obj, TypeVar):
-                    continue
                 obj_module = inspect.getmodule(obj)
                 declared_in_this_module = self.obj.__name__ == _safe_getattr(
                     obj_module, "__name__", None
                 )
-                if declared_in_this_module or name in self._documented_members:
+                include_in_docs = name in self._documented_members or (
+                    declared_in_this_module and not isinstance(obj, TypeVar)
+                )
+                if include_in_docs:
                     members[name] = obj
             for name in self._var_docstrings:
                 members.setdefault(name, empty)
@@ -479,21 +522,21 @@ class Module(Namespace[types.ModuleType]):
         return members
 
     @cached_property
-    def variables(self) -> list["Variable"]:
+    def variables(self) -> list[Variable]:
         """
         A list of all documented module level variables.
         """
         return [x for x in self.members.values() if isinstance(x, Variable)]
 
     @cached_property
-    def classes(self) -> list["Class"]:
+    def classes(self) -> list[Class]:
         """
         A list of all documented module level classes.
         """
         return [x for x in self.members.values() if isinstance(x, Class)]
 
     @cached_property
-    def functions(self) -> list["Function"]:
+    def functions(self) -> list[Function]:
         """
         A list of all documented module level functions.
         """
@@ -510,6 +553,15 @@ class Class(Namespace[type]):
     def __repr__(self):
         return f"<{_decorators(self)}class {self.modulename}.{self.qualname}{_docstr(self)}{_children(self)}>"
 
+    @cached_property
+    def docstring(self) -> str:
+        doc = Doc.docstring.__get__(self)  # type: ignore
+        if doc == dict.__doc__:
+            # Don't display default docstring for dict subclasses (primarily TypedDict).
+            return ""
+        else:
+            return doc
+
     @cached_property
     def _var_docstrings(self) -> dict[str, str]:
         docstrings: dict[str, str] = {}
@@ -518,6 +570,37 @@ class Class(Namespace[type]):
                 docstrings.setdefault(name, docstr)
         return docstrings
 
+    @cached_property
+    def _var_annotations(self) -> dict[str, type]:
+        # this is a bit tricky: __annotations__ also includes annotations from parent classes,
+        # but we need to execute them in the namespace of the parent class.
+        # Our workaround for this is to walk the MRO backwards, and only update/evaluate only if the annotation changes.
+        annotations: dict[
+            str, tuple[Any, type]
+        ] = {}  # attribute -> (annotation_unresolved, annotation_resolved)
+        for cls in reversed(self.obj.__mro__):
+            cls_annotations = doc_ast.walk_tree(cls).annotations.copy()
+            dynamic_annotations = _safe_getattr(cls, "__annotations__", None)
+            if isinstance(dynamic_annotations, dict):
+                for attr, unresolved_annotation in dynamic_annotations.items():
+                    cls_annotations[attr] = unresolved_annotation
+            cls_fullname = (
+                _safe_getattr(cls, "__module__", "") + "." + cls.__qualname__
+            ).lstrip(".")
+
+            new_annotations = {
+                attr: unresolved_annotation
+                for attr, unresolved_annotation in cls_annotations.items()
+                if attr not in annotations
+                or annotations[attr][0] is not unresolved_annotation
+            }
+            for attr, t in resolve_annotations(
+                new_annotations, inspect.getmodule(cls), cls_fullname
+            ).items():
+                annotations[attr] = (new_annotations[attr], t)
+
+        return {k: v[1] for k, v in annotations.items()}
+
     @cached_property
     def _declarations(self) -> dict[str, tuple[str, str]]:
         decls: dict[str, tuple[str, str]] = {}
@@ -538,31 +621,12 @@ class Class(Namespace[type]):
         try:
             return self._declarations[member_name]
         except KeyError:  # pragma: no cover
+            # TypedDict botches __mro__ and may need special casing here.
             warnings.warn(
                 f"Cannot determine where {self.fullname}.{member_name} is taken from, assuming current file."
             )
             return self.modulename, f"{self.qualname}.{member_name}"
 
-    @cached_property
-    def _var_annotations(self) -> dict[str, type]:
-        annotations: dict[str, type] = {}
-        for cls in self.obj.__mro__:
-            cls_annotations = doc_ast.walk_tree(cls).annotations.copy()
-            dynamic_annotations = _safe_getattr(cls, "__annotations__", None)
-            if isinstance(dynamic_annotations, dict):
-                for k, v in dynamic_annotations.items():
-                    cls_annotations[k] = v
-            cls_fullname = (
-                _safe_getattr(cls, "__module__", "") + "." + cls.__qualname__
-            ).lstrip(".")
-            cls_annotations = resolve_annotations(
-                cls_annotations, inspect.getmodule(cls), cls_fullname
-            )
-
-            for k, v in cls_annotations.items():
-                annotations.setdefault(k, v)
-        return annotations
-
     @cached_property
     def own_members(self) -> list[Doc]:
         members = self._members_by_origin.get((self.modulename, self.qualname), [])
@@ -581,32 +645,43 @@ class Class(Namespace[type]):
         for name in self._var_docstrings:
             unsorted.setdefault(name, empty)
 
-        # Do a dance similar to
-        # https://github.com/python/cpython/blob/9feae41c4f04ca27fd2c865807a5caeb50bf4fc4/Lib/inspect.py#L2359-L2376
-        call = _safe_getattr(type(self.obj), "__call__", object.__call__)
-        if not isinstance(call, NonUserDefinedCallables):
-            # Does the metaclass define a custom __call__ method?
-            unsorted["__init__"] = call
-            if issubclass(self.obj, enum.Enum):
-                # Special case: do not show a constructor for enums. They are typically not constructed by users.
+        init_has_no_doc = unsorted.get("__init__", object.__init__).__doc__ in (
+            None,
+            object.__init__.__doc__,
+        )
+        if init_has_no_doc:
+            if inspect.isabstract(self.obj):
+                # Special case: We don't want to show constructors for abstract base classes unless
+                # they have a custom docstring.
+                del unsorted["__init__"]
+            elif issubclass(self.obj, enum.Enum):
+                # Special case: Do not show a constructor for enums. They are typically not constructed by users.
                 # The alternative would be showing __new__, as __call__ is too verbose.
                 del unsorted["__init__"]
-        else:
-            # Does our class define a custom __new__ method?
-            new = _safe_getattr(self.obj, "__new__", object.__new__)
-            if not isinstance(new, NonUserDefinedCallables):
-                # we only want to pick up __new__ if it has a non-default docstring.
-                prefer_init_over_new = new.__doc__ in (None, object.__new__.__doc__)
-                if not prefer_init_over_new:
-                    unsorted["__init__"] = new
-
-        if inspect.isabstract(self.obj) and _safe_getattr(
-            self.obj, "__init__", object.__init__
-        ).__doc__ in (None, object.__init__.__doc__):
-            # Special case: We don't want to show constructors for abstract base classes unless
-            # they have a custom docstring. If a docstring is present and display is still undesired,
-            # users need to override the is_public macro.
-            del unsorted["__init__"]
+            elif issubclass(self.obj, dict):
+                # Special case: Do not show a constructor for dict subclasses.
+                del unsorted["__init__"]
+            else:
+                # Check if there's a helpful Metaclass.__call__ or Class.__new__. This dance is very similar to
+                # https://github.com/python/cpython/blob/9feae41c4f04ca27fd2c865807a5caeb50bf4fc4/Lib/inspect.py#L2359-L2376
+                call = _safe_getattr(type(self.obj), "__call__", None)
+                custom_call_with_custom_docstring = (
+                    call is not None
+                    and not isinstance(call, NonUserDefinedCallables)
+                    and call.__doc__ not in (None, object.__call__.__doc__)
+                )
+                if custom_call_with_custom_docstring:
+                    unsorted["__init__"] = call
+                else:
+                    # Does our class define a custom __new__ method?
+                    new = _safe_getattr(self.obj, "__new__", None)
+                    custom_new_with_custom_docstring = (
+                        new is not None
+                        and not isinstance(new, NonUserDefinedCallables)
+                        and new.__doc__ not in (None, object.__new__.__doc__)
+                    )
+                    if custom_new_with_custom_docstring:
+                        unsorted["__init__"] = new
 
         sorted: dict[str, Any] = {}
         for cls in self.obj.__mro__:
@@ -645,7 +720,7 @@ class Class(Namespace[type]):
         return decorators
 
     @cached_property
-    def class_variables(self) -> list["Variable"]:
+    def class_variables(self) -> list[Variable]:
         """
         A list of all documented class variables in the class.
 
@@ -659,7 +734,7 @@ class Class(Namespace[type]):
         ]
 
     @cached_property
-    def instance_variables(self) -> list["Variable"]:
+    def instance_variables(self) -> list[Variable]:
         """
         A list of all instance variables in the class.
         """
@@ -670,7 +745,7 @@ class Class(Namespace[type]):
         ]
 
     @cached_property
-    def classmethods(self) -> list["Function"]:
+    def classmethods(self) -> list[Function]:
         """
         A list of all documented `@classmethod`s.
         """
@@ -681,7 +756,7 @@ class Class(Namespace[type]):
         ]
 
     @cached_property
-    def staticmethods(self) -> list["Function"]:
+    def staticmethods(self) -> list[Function]:
         """
         A list of all documented `@staticmethod`s.
         """
@@ -692,7 +767,7 @@ class Class(Namespace[type]):
         ]
 
     @cached_property
-    def methods(self) -> list["Function"]:
+    def methods(self) -> list[Function]:
         """
         A list of all documented methods in the class that are neither static- nor classmethods.
         """
@@ -705,7 +780,10 @@ class Class(Namespace[type]):
         ]
 
 
-WrappedFunction = Union[types.FunctionType, staticmethod, classmethod]
+if sys.version_info >= (3, 10):
+    WrappedFunction = types.FunctionType | staticmethod | classmethod
+else:  # pragma: no cover
+    WrappedFunction = Union[types.FunctionType, staticmethod, classmethod]
 
 
 class Function(Doc[types.FunctionType]):
@@ -716,7 +794,7 @@ class Function(Doc[types.FunctionType]):
     supports `@classmethod`s or `@staticmethod`s.
     """
 
-    wrapped: Union[types.FunctionType, staticmethod, classmethod]
+    wrapped: WrappedFunction
     """The original wrapped function (e.g., `staticmethod(func)`)"""
 
     obj: types.FunctionType
@@ -856,9 +934,7 @@ class Variable(Doc[None]):
     Representation of a variable's documentation. This includes module, class and instance variables.
     """
 
-    default_value: Union[
-        Any, empty
-    ]  # technically Any includes empty, but this conveys intent.
+    default_value: Any | empty  # technically Any includes empty, but this conveys intent.
     """
     The variable's default value.
     
@@ -867,7 +943,7 @@ class Variable(Doc[None]):
     To distinguish this case from a default value of `None`, `pdoc.doc_types.empty` is used as a placeholder.
     """
 
-    annotation: Union[type, empty]
+    annotation: type | empty
     """
     The variable's type annotation.
     
@@ -881,8 +957,8 @@ class Variable(Doc[None]):
         *,
         taken_from: tuple[str, str],
         docstring: str,
-        annotation: Union[type, empty] = empty,
-        default_value: Union[Any, empty] = empty,
+        annotation: type | empty = empty,
+        default_value: Any | empty = empty,
     ):
         """
         Construct a variable doc object.
@@ -916,12 +992,10 @@ class Variable(Doc[None]):
         """The variable's default value as a pretty-printed str."""
         if self.default_value is empty:
             return ""
-        elif isinstance(self.default_value, _tuplegetter):
-            return ""
         else:
             try:
                 return re.sub(
-                    r" at 0x[0-9a-fA-F]+(?=>$)",
+                    r" at 0x[0-9a-fA-F]+(?=>)",
                     "",
                     f" = {repr(self.default_value)}",
                 )
@@ -943,14 +1017,15 @@ class _PrettySignature(inspect.Signature):
     for complex signatures.
     """
 
-    def __str__(self):
+    MULTILINE_CUTOFF = 70
+
+    def _params(self) -> list[str]:
         # redeclared here to keep code snipped below as-is.
         _POSITIONAL_ONLY = inspect.Parameter.POSITIONAL_ONLY
         _VAR_POSITIONAL = inspect.Parameter.VAR_POSITIONAL
         _KEYWORD_ONLY = inspect.Parameter.KEYWORD_ONLY
-        _empty = empty
 
-        # https://github.com/python/cpython/blob/799f8489d418b7f9207d333eac38214931bd7dcc/Lib/inspect.py#L3083-L3123
+        # https://github.com/python/cpython/blob/799f8489d418b7f9207d333eac38214931bd7dcc/Lib/inspect.py#L3083-L3117
         # Change: added re.sub() to formatted = ....
         # ✂ start ✂
         result = []
@@ -988,18 +1063,28 @@ class _PrettySignature(inspect.Signature):
             # There were only positional-only parameters, hence the
             # flag was not reset to 'False'
             result.append("/")
-
-        rendered = "({})".format(", ".join(result))
-
-        if self.return_annotation is not _empty:
-            anno = formatannotation(self.return_annotation)
-            rendered += " -> {}".format(anno)
         # ✂ end ✂
 
-        if len(rendered) > 70:
+        return result
+
+    def _return_annotation_str(self) -> str:
+        if self.return_annotation is not empty:
+            return formatannotation(self.return_annotation)
+        else:
+            return ""
+
+    def __str__(self):
+        result = self._params()
+        return_annot = self._return_annotation_str()
+
+        total_len = sum(len(x) + 2 for x in result) + len(return_annot)
+
+        if total_len > self.MULTILINE_CUTOFF:
             rendered = "(\n    " + ",\n    ".join(result) + "\n)"
-            if self.return_annotation is not _empty:
-                rendered += f" -> {formatannotation(self.return_annotation)}"
+        else:
+            rendered = "({})".format(", ".join(result))
+        if return_annot:
+            rendered += f" -> {return_annot}"
 
         return rendered
 
@@ -1012,7 +1097,7 @@ def _cut(x: str) -> str:
         return x[:20] + "…"
 
 
-def _docstr(doc: "Doc") -> str:
+def _docstr(doc: Doc) -> str:
     """helper function for Doc.__repr__()"""
     docstr = []
     if doc.is_inherited:
@@ -1025,7 +1110,7 @@ def _docstr(doc: "Doc") -> str:
         return ""
 
 
-def _decorators(doc: Union["Class", "Function"]) -> str:
+def _decorators(doc: Class | Function) -> str:
     """helper function for Doc.__repr__()"""
     if doc.decorators:
         return " ".join(doc.decorators) + " "
@@ -1050,8 +1135,7 @@ def _safe_getattr(obj, attr, default):
         return getattr(obj, attr, default)
     except Exception as e:
         warnings.warn(
-            f"getattr({obj!r}, {attr!r}, {default!r}) raised an exception: {e}",
-            RuntimeWarning,
+            f"getattr({obj!r}, {attr!r}, {default!r}) raised an exception: {e!r}"
         )
         return default
 
@@ -1061,10 +1145,7 @@ def _safe_getdoc(obj: Any) -> str:
     try:
         doc = inspect.getdoc(obj) or ""
     except Exception as e:
-        warnings.warn(
-            f"inspect.getdoc({obj!r}) raised an exception: {e}",
-            RuntimeWarning,
-        )
+        warnings.warn(f"inspect.getdoc({obj!r}) raised an exception: {e!r}")
         return ""
     else:
         return doc.strip()
