@@ -9,9 +9,10 @@ import ast
 import inspect
 import types
 import warnings
+from collections.abc import Iterable, Iterator
 from dataclasses import dataclass
 from itertools import tee, zip_longest
-from typing import Any, Iterable, Iterator, TypeVar, Union, overload
+from typing import Any, TypeVar, overload
 
 from ._compat import ast_unparse, cache
 
@@ -44,7 +45,7 @@ def parse(obj: types.ModuleType) -> ast.Module:
 
 
 @overload
-def parse(obj: types.FunctionType) -> Union[ast.FunctionDef, ast.AsyncFunctionDef]:
+def parse(obj: types.FunctionType) -> ast.FunctionDef | ast.AsyncFunctionDef:
     ...
 
 
@@ -86,7 +87,7 @@ class AstInfo:
     Annotations are not evaluated by this module and only returned as strings."""
 
 
-def walk_tree(obj: Union[types.ModuleType, type]) -> AstInfo:
+def walk_tree(obj: types.ModuleType | type) -> AstInfo:
     """
     Walks the abstract syntax tree for `obj` and returns the extracted information.
     """
@@ -95,7 +96,7 @@ def walk_tree(obj: Union[types.ModuleType, type]) -> AstInfo:
 
 @cache
 def _walk_tree(
-    tree: Union[ast.Module, ast.ClassDef, ast.FunctionDef, ast.AsyncFunctionDef]
+    tree: ast.Module | ast.ClassDef | ast.FunctionDef | ast.AsyncFunctionDef,
 ) -> AstInfo:
     docstrings = {}
     annotations = {}
@@ -132,13 +133,13 @@ T = TypeVar("T")
 
 
 def sort_by_source(
-    obj: Union[types.ModuleType, type], sorted: dict[str, T], unsorted: dict[str, T]
+    obj: types.ModuleType | type, sorted: dict[str, T], unsorted: dict[str, T]
 ) -> tuple[dict[str, T], dict[str, T]]:
     """
     Takes items from `unsorted` and inserts them into `sorted` in order of appearance in the source code of `obj`.
     The only exception to this rule is `__init__`, which (if present) is always inserted first.
 
-    Some items may not be found, for example because they've been inherited from a superclass. The are returned as-is.
+    Some items may not be found, for example because they've been inherited from a superclass. They are returned as-is.
 
     Returns a `(sorted, not found)` tuple.
     """
@@ -185,7 +186,8 @@ def type_checking_sections(mod: types.ModuleType) -> ast.Module:
             isinstance(node, ast.If)
             and isinstance(node.test, ast.Attribute)
             and isinstance(node.test.value, ast.Name)
-            and node.test.value.id == "typing"
+            # some folks do "import typing as t", the accuracy with just TYPE_CHECKING is good enough.
+            # and node.test.value.id == "typing"
             and node.test.attr == "TYPE_CHECKING"
         ):
             ret.body.extend(node.body)
@@ -221,7 +223,7 @@ def _parse_class(source: str) -> ast.ClassDef:
 
 
 @cache
-def _parse_function(source: str) -> Union[ast.FunctionDef, ast.AsyncFunctionDef]:
+def _parse_function(source: str) -> ast.FunctionDef | ast.AsyncFunctionDef:
     """
     Parse the AST for the source code of a (async) function and return the matching AST node.
 
@@ -242,14 +244,11 @@ def _parse_function(source: str) -> Union[ast.FunctionDef, ast.AsyncFunctionDef]
 
 def _parse(
     source: str,
-) -> Union[ast.Module, ast.ClassDef, ast.FunctionDef, ast.AsyncFunctionDef]:
+) -> ast.Module | ast.ClassDef | ast.FunctionDef | ast.AsyncFunctionDef:
     try:
         return ast.parse(_dedent(source))
     except Exception as e:
-        warnings.warn(
-            f"Error parsing source code: {e}\n" f"===\n" f"{source}\n" f"===",
-            RuntimeWarning,
-        )
+        warnings.warn(f"Error parsing source code: {e}\n" f"===\n" f"{source}\n" f"===")
         return ast.parse("")
 
 
@@ -280,7 +279,7 @@ def _dedent(source: str) -> str:
 
 
 @cache
-def _nodes(tree: Union[ast.Module, ast.ClassDef]) -> list[ast.AST]:
+def _nodes(tree: ast.Module | ast.ClassDef) -> list[ast.AST]:
     """
     Returns the list of all nodes in tree's body, but also inlines the body of __init__.
 
@@ -289,7 +288,7 @@ def _nodes(tree: Union[ast.Module, ast.ClassDef]) -> list[ast.AST]:
     return list(_nodes_iter(tree))
 
 
-def _nodes_iter(tree: Union[ast.Module, ast.ClassDef]) -> Iterator[ast.AST]:
+def _nodes_iter(tree: ast.Module | ast.ClassDef) -> Iterator[ast.AST]:
     for a in tree.body:
         yield a
         if isinstance(a, ast.FunctionDef) and a.name == "__init__":
