@@ -12,20 +12,24 @@ import linecache
 import os
 import pkgutil
 import platform
+import re
+import shutil
 import subprocess
 import sys
 import traceback
 import types
 import warnings
+from collections.abc import Callable, Iterable, Iterator, Sequence
 from contextlib import contextmanager
 from pathlib import Path
-from typing import Callable, Iterable, Iterator, Optional, Sequence, Union
 from unittest.mock import patch
 
-from . import doc_ast, docstrings
+import pdoc
+import pdoc.doc_ast
+import pdoc.docstrings
 
 
-def walk_specs(specs: Sequence[Union[Path, str]]) -> dict[str, None]:
+def walk_specs(specs: Sequence[Path | str]) -> list[str]:
     """
     This function processes a list of module specifications and returns a collection of module names, including all
     submodules, that should be processed by pdoc.
@@ -38,14 +42,18 @@ def walk_specs(specs: Sequence[Union[Path, str]]) -> dict[str, None]:
      - `./test/testdata/demo_long.py`
      - `./test/testdata/demopackage`
 
-
-    Practically you can think of this function returning a list. Technically we return a dict with empty values,
-    which has efficient `__iter__` and `__contains__` implementations.
-
-    *This function has side-effects:* See `parse_spec`.
+    *This function has side effects:* See `parse_spec`.
     """
     all_modules: dict[str, None] = {}
     for spec in specs:
+
+        if isinstance(spec, str) and spec.startswith("!"):
+            ignore_pattern = re.compile(spec[1:])
+            all_modules = {
+                k: v for k, v in all_modules.items() if not ignore_pattern.match(k)
+            }
+            continue
+
         modname = parse_spec(spec)
 
         try:
@@ -56,7 +64,6 @@ def walk_specs(specs: Sequence[Union[Path, str]]) -> dict[str, None]:
         except AnyException:
             warnings.warn(
                 f"Cannot find spec for {modname} (from {spec}):\n{traceback.format_exc()}",
-                RuntimeWarning,
                 stacklevel=2,
             )
         else:
@@ -66,21 +73,30 @@ def walk_specs(specs: Sequence[Union[Path, str]]) -> dict[str, None]:
                 ispkg=bool(modspec.submodule_search_locations),
             )
             for m in walk_packages2([mod_info]):
+                if m.name in all_modules:
+                    warnings.warn(
+                        f"The module specification {spec!r} adds a module named {m.name}, but a module with this name "
+                        f"has already been added. You may have accidentally repeated a module spec, or you are trying "
+                        f"to document two modules with the same filename from two different directories, which does "
+                        f"not work. Only one documentation page will be generated."
+                    )
                 all_modules[m.name] = None
 
     if not all_modules:
-        raise ValueError(f"Module not found: {', '.join(str(x) for x in specs)}.")
+        raise ValueError(
+            f"No modules found matching spec: {', '.join(str(x) for x in specs)}"
+        )
 
-    return all_modules
+    return list(all_modules)
 
 
-def parse_spec(spec: Union[Path, str]) -> str:
+def parse_spec(spec: Path | str) -> str:
     """
     This functions parses a user's module specification into a module identifier that can be imported.
     If both a local file/directory and an importable module with the same name exist, a warning will be printed.
 
-    *This function has side-effects:* `sys.path` will be amended if the specification is a path.
-    If this side-effect is undesired, pass a module name instead.
+    *This function has side effects:* `sys.path` will be amended if the specification is a path.
+    If this side effect is undesired, pass a module name instead.
     """
     pspec = Path(spec)
     if isinstance(spec, str) and (os.sep in spec or (os.altsep and os.altsep in spec)):
@@ -105,57 +121,75 @@ def parse_spec(spec: Union[Path, str]) -> str:
             )
             local_dir = Path(spec).absolute()
             if local_dir not in (origin, origin.parent):
-                print(
-                    f"Warning: {spec!r} may refer to either the installed Python module or the local file/directory "
-                    f"with the same name. pdoc will document the installed module, prepend './' to force "
-                    f"documentation of the local file/directory.\n"
+                warnings.warn(
+                    f"{spec!r} may refer to either the installed Python module or the local file/directory with the "
+                    f"same name. pdoc will document the installed module, prepend './' to force documentation of the "
+                    f"local file/directory.\n"
                     f" - Module location: {origin}\n"
                     f" - Local file/directory: {local_dir}",
-                    file=sys.stderr,
+                    RuntimeWarning,
                 )
 
     if isinstance(spec, Path):
         if (spec.parent / "__init__.py").exists():
             return parse_spec(spec.resolve().parent) + f".{spec.stem}"
-        if str(spec.parent) not in sys.path:
-            sys.path.insert(0, str(spec.parent))
-        if spec.stem in sys.modules:
+        parent_dir = str(spec.parent)
+        sys.path = [parent_dir] + [x for x in sys.path if x != parent_dir]
+        if spec.stem in sys.modules and sys.modules[spec.stem].__file__:
             local_dir = spec.resolve()
-            origin = Path(sys.modules[spec.stem].__file__).resolve()
+            file = sys.modules[spec.stem].__file__
+            assert file is not None  # make mypy happy
+            origin = Path(file).resolve()
             if local_dir not in (origin, origin.parent, origin.with_suffix("")):
-                print(
-                    f"Warning: pdoc cannot load {spec.stem!r} because a module with the same name is already "
-                    f"imported in pdoc's Python process. pdoc will document the loaded module from {origin} instead.",
-                    file=sys.stderr,
+                warnings.warn(
+                    f"pdoc cannot load {spec.stem!r} because a module with the same name is already imported in pdoc's "
+                    f"Python process. pdoc will document the loaded module from {origin} instead.",
+                    RuntimeWarning,
                 )
         return spec.stem
     else:
         return spec
 
 
+def _noop(*args, **kwargs):
+    pass
+
+
+class PdocDefusedPopen(subprocess.Popen):
+    if platform.system() == "Windows":  # pragma: no cover
+        _noop_exe = "echo.exe"
+    else:  # pragma: no cover
+        _noop_exe = "echo"
+
+    def __init__(self, *args, **kwargs):  # pragma: no cover
+        command_allowed = (
+            args
+            and args[0]
+            and args[0][0]
+            in (
+                # these invocations may all come from https://github.com/python/cpython/blob/main/Lib/ctypes/util.py,
+                # which we want to keep working.
+                "/sbin/ldconfig",
+                "ld",
+                shutil.which("gcc") or shutil.which("cc"),
+                shutil.which("objdump"),
+            )
+        )
+        if not command_allowed:
+            kwargs["executable"] = self._noop_exe
+        super().__init__(*args, **kwargs)
+
+
 @contextmanager
 def mock_some_common_side_effects():
     """
     This context manager is applied when importing modules. It mocks some common side effects that may happen upon
-    module import. For example, `import antigravity` normally causes a webbrowser to open, which we want to suppress.
+    module import. For example, `import antigravity` normally causes a web browser to open, which we want to suppress.
 
     Note that this function must not be used for security purposes, it's easily bypassable.
     """
-    if platform.system() == "Windows":  # pragma: no cover
-        noop_exe = "echo.exe"
-    else:  # pragma: no cover
-        noop_exe = "echo"
-
-    def noop(*args, **kwargs):
-        pass
-
-    class PdocDefusedPopen(subprocess.Popen):
-        def __init__(self, *args, **kwargs):  # pragma: no cover
-            kwargs["executable"] = noop_exe
-            super().__init__(*args, **kwargs)
-
     with patch("subprocess.Popen", new=PdocDefusedPopen), patch(
-        "os.startfile", new=noop, create=True
+        "os.startfile", new=_noop, create=True
     ), patch("sys.stdout", new=io.StringIO()), patch(
         "sys.stderr", new=io.StringIO()
     ), patch(
@@ -215,13 +249,10 @@ def walk_packages2(
             try:
                 module = load_module(mod.name)
             except RuntimeError:
-                warnings.warn(
-                    f"Error loading {mod.name}:\n{traceback.format_exc()}",
-                    RuntimeWarning,
-                )
+                warnings.warn(f"Error loading {mod.name}:\n{traceback.format_exc()}")
                 continue
 
-            mod_all: list[str] = getattr(module, "__all__", None)
+            mod_all = getattr(module, "__all__", None)
             if mod_all is not None:
                 filt = mod_all.__contains__
             else:
@@ -233,7 +264,7 @@ def walk_packages2(
             yield from walk_packages2(pkgutil.iter_modules(path, f"{mod.name}."), filt)
 
 
-def module_mtime(modulename: str) -> Optional[float]:
+def module_mtime(modulename: str) -> float | None:
     """Returns the time the specified module file was last modified, or `None` if this cannot be determined.
     The primary use of this is live-reloading modules on modification."""
     try:
@@ -251,7 +282,7 @@ def invalidate_caches(module_name: str) -> None:
     """
     Invalidate module cache to allow live-reloading of modules.
     """
-    # Getting this right is tricky – reloading modules causes a bunch of surprising side-effects.
+    # Getting this right is tricky – reloading modules causes a bunch of surprising side effects.
     # Our current best effort is to call `importlib.reload` on all modules that start with module_name.
     # We also exclude our own dependencies, which cause fun errors otherwise.
     if module_name not in sys.modules:
@@ -272,8 +303,9 @@ def invalidate_caches(module_name: str) -> None:
 
     importlib.invalidate_caches()
     linecache.clearcache()
-    doc_ast._get_source.cache_clear()
-    docstrings.convert.cache_clear()
+    pdoc.doc.Module.from_name.cache_clear()
+    pdoc.doc_ast._get_source.cache_clear()
+    pdoc.docstrings.convert.cache_clear()
 
     prefix = f"{module_name}."
     mods = sorted(
@@ -291,17 +323,5 @@ def invalidate_caches(module_name: str) -> None:
         except AnyException:
             warnings.warn(
                 f"Error reloading {modname}:\n{traceback.format_exc()}",
-                RuntimeWarning,
                 stacklevel=2,
             )
-
-
-def parse_specs(
-    modules: Sequence[Union[Path, str]]
-) -> dict[str, None]:  # pragma: no cover
-    """A deprecated alias for `walk_specs`."""
-    warnings.warn(
-        "pdoc.extract.parse_specs has been renamed to pdoc.extract.walk_specs",
-        PendingDeprecationWarning,
-    )
-    return walk_specs(modules)
